# Copyright (c) 2023 ING Analytics Wholesale Banking
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

from popmon.alerting import (
    AlertsSummary,
    DynamicBounds,
    StaticBounds,
    TrafficLightAlerts,
    traffic_light_summary,
)
from popmon.analysis.apply_func import ApplyFunc
from popmon.analysis.comparison.hist_comparer import (
    ExpandingHistComparer,
    PreviousHistComparer,
    ReferenceHistComparer,
    RollingHistComparer,
)
from popmon.analysis.functions import rolling_lr_zscore
from popmon.analysis.profiling import HistProfiler
from popmon.analysis.profiling.pull_calculator import (
    ExpandingPullCalculator,
    ReferencePullCalculator,
    RefMedianMadPullCalculator,
    RollingPullCalculator,
)
from popmon.base import Module, Pipeline
from popmon.config import Settings
from popmon.hist.hist_splitter import HistSplitter

from .timing import Timing


def get_metrics_pipeline_class(reference_type, reference):
    _metrics_pipeline_register = {
        "self": SelfReferenceMetricsPipeline,
        "external": ExternalReferenceMetricsPipeline,
        "self_split": ExternalReferenceMetricsPipeline,
        "rolling": RollingReferenceMetricsPipeline,
        "expanding": ExpandingReferenceMetricsPipeline,
    }

    if reference_type not in _metrics_pipeline_register:
        raise ValueError(
            f"reference_type should be in {_metrics_pipeline_register.keys()!s}'."
        )
    if (
        reference_type in ["external", "self_split"]
        and not isinstance(reference, dict)
        and reference is not None
    ):
        raise TypeError("reference should be a dict of histogrammar histograms.")

    return _metrics_pipeline_register[reference_type]


def create_metrics_pipeline(
    settings: Settings,
    reference=None,
    hists_key: str = "hists",
    **kwargs,
):
    # configuration and datastore for report pipeline
    cfg = {
        "hists_key": hists_key,
        "settings": settings,
        **kwargs,
    }

    # execute reporting pipeline
    cls = get_metrics_pipeline_class(settings.reference_type, reference)
    pipeline = cls(**cfg)
    return pipeline


def get_timing_module(key: str) -> list[Module]:
    return [Timing(store_key=key)]


def get_splitting_modules(hists_key, features, time_axis) -> list[Module | Pipeline]:
    """
    Splitting of test histograms. For each histogram with datetime i, comparison of histogram i with histogram i-1,
    results in chi2 comparison of histograms
    """
    modules: list[Module | Pipeline] = [
        HistSplitter(
            read_key=hists_key,
            store_key="split_hists",
            features=features,
            feature_begins_with=f"{time_axis}:",
        ),
        PreviousHistComparer(read_key="split_hists", store_key="comparisons"),
        HistProfiler(read_key="split_hists", store_key="profiles"),
    ]
    return modules


def get_traffic_light_modules(monitoring_rules) -> list[Module | Pipeline]:
    """
    Expand all (wildcard) static traffic light bounds and apply them.
    Applied to both profiles and comparisons datasets
    """
    modules: list[Module | Pipeline] = [
        TrafficLightAlerts(
            read_key="profiles",
            rules=monitoring_rules,
            store_key="traffic_lights",
            expanded_rules_key="static_bounds",
        ),
        TrafficLightAlerts(
            read_key="comparisons",
            rules=monitoring_rules,
            store_key="traffic_lights",
            expanded_rules_key="static_bounds_comparisons",
        ),
        ApplyFunc(
            apply_to_key="traffic_lights",
            apply_funcs=[{"func": traffic_light_summary, "axis": 1, "suffix": ""}],
            assign_to_key="alerts",
            msg="Generating traffic light alerts summary.",
        ),
        AlertsSummary(read_key="alerts"),
    ]
    return modules


def get_static_bound_modules(pull_rules) -> list[Module | Pipeline]:
    """
    generate dynamic traffic light boundaries, based on traffic lights for normalized residuals, used for
    plotting in popmon_profiles report.
    """
    modules: list[Module | Pipeline] = [
        StaticBounds(
            read_key="profiles",
            rules=pull_rules,
            store_key="dynamic_bounds",
            suffix_mean="_mean",
            suffix_std="_std",
        ),
        StaticBounds(
            read_key="comparisons",
            rules=pull_rules,
            store_key="dynamic_bounds_comparisons",
            suffix_mean="_mean",
            suffix_std="_std",
        ),
    ]
    return modules


def get_dynamic_bound_modules(pull_rules) -> list[Module | Pipeline]:
    """
    Generate dynamic traffic light boundaries, based on traffic lights for normalized residuals, used for
    plotting in popmon_profiles report.
    """
    modules: list[Module | Pipeline] = [
        DynamicBounds(
            read_key="profiles",
            rules=pull_rules,
            store_key="dynamic_bounds",
            suffix_mean="_mean",
            suffix_std="_std",
        ),
        DynamicBounds(
            read_key="comparisons",
            rules=pull_rules,
            store_key="dynamic_bounds_comparisons",
            suffix_mean="_mean",
            suffix_std="_std",
        ),
    ]
    return modules


def get_trend_modules(window) -> list[Module | Pipeline]:
    """Looking for significant rolling linear trends in selected features/metrics"""
    modules: list[Module | Pipeline] = [
        ApplyFunc(
            apply_to_key="profiles",
            assign_to_key="comparisons",
            apply_funcs=[
                {
                    "func": rolling_lr_zscore,
                    "suffix": f"_trend{window}_zscore",
                    "entire": True,
                    "window": window,
                    "metrics": ["mean", "phik", "fraction_true"],
                }
            ],
            msg="Computing significance of (rolling) trend in means of features",
        ),
    ]
    return modules


class SelfReferenceMetricsPipeline(Pipeline):
    def __init__(
        self,
        settings: Settings,
        hists_key,
    ) -> None:
        """Example metrics pipeline for comparing test data with itself (full test set)

        :param str hists_key: key to test histograms in datastore. default is 'test_hists'
        :return: assembled self reference pipeline
        """
        from popmon.analysis.comparison import Comparisons

        reference_prefix = "ref"
        reference_modules: list[Module | Pipeline] = [
            # 3. Comparison of with profiled test histograms, results in chi2 comparison of histograms
            ReferenceHistComparer(
                reference_key="split_hists",
                assign_to_key="split_hists",
                store_key="comparisons",
            ),
            RefMedianMadPullCalculator(
                reference_key="comparisons",
                assign_to_key="comparisons",
                suffix_mean="_mean",
                suffix_std="_std",
                suffix_pull="_pull",
                metrics=[
                    f"{reference_prefix}_{key}"
                    for key in Comparisons.get_keys()
                    if key in ["max_prob_diff", "psi", "jsd"]
                ],
            ),
            # 4. profiling of histograms, then pull calculation compared with reference mean and std,
            #        to obtain normalized residuals of profiles
            RefMedianMadPullCalculator(
                reference_key="profiles",
                assign_to_key="profiles",
                suffix_mean="_mean",
                suffix_std="_std",
                suffix_pull="_pull",
            ),
        ]

        modules = (
            get_timing_module("start_time")
            + get_splitting_modules(hists_key, settings.features, settings.time_axis)
            + reference_modules
            + get_trend_modules(settings.comparison.window)
            + get_static_bound_modules(settings.monitoring.pull_rules)
            + get_traffic_light_modules(settings.monitoring.monitoring_rules)
            + get_timing_module("end_time")
        )
        super().__init__(modules)


class ExternalReferenceMetricsPipeline(Pipeline):
    def __init__(
        self,
        settings: Settings,
        hists_key: str = "test_hists",
        ref_hists_key: str = "ref_hists",
    ) -> None:
        """Example metrics pipeline for comparing test data with other (full) external reference set

        :param str hists_key: key to test histograms in datastore. default is 'test_hists'
        :param str ref_hists_key: key to reference histograms in datastore. default is 'ref_hists'
        :return: assembled external reference pipeline
        """
        from popmon.analysis.comparison import Comparisons

        reference_prefix = "ref"
        reference_modules: list[Module | Pipeline] = [
            # 3. Profiling of split reference histograms, then chi2 comparison with test histograms
            HistSplitter(
                read_key=ref_hists_key,
                store_key="split_ref_hists",
                features=settings.features,
                feature_begins_with=f"{settings.time_axis}:",
            ),
            ReferenceHistComparer(
                reference_key="split_ref_hists",
                assign_to_key="split_hists",
                store_key="comparisons",
            ),
            HistProfiler(read_key="split_ref_hists", store_key="ref_profiles"),
            RefMedianMadPullCalculator(
                reference_key="comparisons",
                assign_to_key="comparisons",
                suffix_mean="_mean",
                suffix_std="_std",
                suffix_pull="_pull",
                metrics=[
                    f"{reference_prefix}_{key}"
                    for key in Comparisons.get_keys()
                    if key in ["max_prob_diff", "psi", "jsd"]
                ],
            ),
            # 4. pull calculation compared with reference mean and std, to obtain normalized residuals of profiles
            ReferencePullCalculator(
                reference_key="ref_profiles",
                assign_to_key="profiles",
                suffix_mean="_mean",
                suffix_std="_std",
                suffix_pull="_pull",
            ),
        ]
        modules = (
            get_timing_module("start_time")
            + get_splitting_modules(hists_key, settings.features, settings.time_axis)
            + reference_modules
            + get_trend_modules(settings.comparison.window)
            + get_static_bound_modules(settings.monitoring.pull_rules)
            + get_traffic_light_modules(settings.monitoring.monitoring_rules)
            + get_timing_module("end_time")
        )
        super().__init__(modules)


class RollingReferenceMetricsPipeline(Pipeline):
    def __init__(
        self,
        settings: Settings,
        hists_key: str = "test_hists",
    ) -> None:
        """Example metrics pipeline for comparing test data with itself (rolling test set)

        :param str hists_key: key to test histograms in datastore. default is 'test_hists'
        :return: assembled rolling reference pipeline
        """
        from popmon.analysis.comparison import Comparisons

        reference_prefix = "roll"
        reference_modules: list[Module | Pipeline] = [
            # 3. profiling of reference histograms, then comparison of with profiled test histograms
            #        results in chi2 comparison of histograms
            RollingHistComparer(
                read_key="split_hists",
                window=settings.comparison.window,
                shift=settings.comparison.shift,
                store_key="comparisons",
            ),
            RefMedianMadPullCalculator(
                reference_key="comparisons",
                assign_to_key="comparisons",
                suffix_mean="_mean",
                suffix_std="_std",
                suffix_pull="_pull",
                metrics=[
                    f"{reference_prefix}_{key}"
                    for key in Comparisons.get_keys()
                    if key in ["max_prob_diff", "psi", "jsd"]
                ],
            ),
            # 4. profiling of histograms, then pull calculation compared with reference mean and std,
            #        to obtain normalized residuals of profiles
            RollingPullCalculator(
                read_key="profiles",
                window=settings.comparison.window,
                shift=settings.comparison.shift,
                suffix_mean="_mean",
                suffix_std="_std",
                suffix_pull="_pull",
            ),
        ]

        modules = (
            get_timing_module("start_time")
            + get_splitting_modules(hists_key, settings.features, settings.time_axis)
            + reference_modules
            + get_trend_modules(settings.comparison.window)
            + get_dynamic_bound_modules(settings.monitoring.pull_rules)
            + get_traffic_light_modules(settings.monitoring.monitoring_rules)
            + get_timing_module("end_time")
        )
        super().__init__(modules)


class ExpandingReferenceMetricsPipeline(Pipeline):
    def __init__(
        self,
        settings: Settings,
        hists_key: str = "test_hists",
    ) -> None:
        """Example metrics pipeline for comparing test data with itself (expanding test set)

        :param str hists_key: key to test histograms in datastore. default is 'test_hists'
        :return: assembled expanding reference pipeline
        """
        from popmon.analysis.comparison import Comparisons

        reference_prefix = "expanding"
        reference_modules: list[Module | Pipeline] = [
            # 3. profiling of reference histograms, then comparison of with profiled test histograms
            #    results in chi2 comparison of histograms
            ExpandingHistComparer(
                read_key="split_hists",
                shift=settings.comparison.shift,
                store_key="comparisons",
            ),
            # 4. profiling of histograms, then pull calculation compared with reference mean and std,
            #        to obtain normalized residuals of profiles
            RefMedianMadPullCalculator(
                reference_key="comparisons",
                assign_to_key="comparisons",
                suffix_mean="_mean",
                suffix_std="_std",
                suffix_pull="_pull",
                metrics=[
                    f"{reference_prefix}_{key}"
                    for key in Comparisons.get_keys()
                    if key in ["max_prob_diff", "psi", "jsd"]
                ],
            ),
            ExpandingPullCalculator(
                read_key="profiles",
                shift=settings.comparison.shift,
                suffix_mean="_mean",
                suffix_std="_std",
                suffix_pull="_pull",
            ),
        ]

        modules = (
            get_timing_module("start_time")
            + get_splitting_modules(hists_key, settings.features, settings.time_axis)
            + reference_modules
            + get_trend_modules(settings.comparison.window)
            + get_dynamic_bound_modules(settings.monitoring.pull_rules)
            + get_traffic_light_modules(settings.monitoring.monitoring_rules)
            + get_timing_module("end_time")
        )
        super().__init__(modules)



# Copyright (c) 2023 ING Analytics Wholesale Banking
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

from collections import defaultdict

import numpy as np
import pandas as pd
from tqdm import tqdm

from popmon.analysis.comparison import Comparisons
from popmon.analysis.profiling import Profiles
from popmon.base import Module
from popmon.config import Report
from popmon.utils import filter_metrics, parallel, short_date
from popmon.visualization.utils import _prune, plot_bars

profiles = Profiles.get_descriptions()

comparisons = Comparisons.get_descriptions()


group_titles = {
    "prev1": "Previous Reference",
    "rolling": "Rolling Reference",
    "self": "Self-Reference",
    "ref": "External Reference",
    "expanding": "Expanding Reference",
}
references = {
    "ref": "the reference data",
    "roll": "a rolling window",
    "prev1": "the preceding time slot",
    "expanding": "all preceding time slots",
}
group_descriptions = {
    key: f"Comparing each time slot to {value}." for key, value in references.items()
}


def get_stat_description(name: str) -> str:
    """Gets the description of a statistic.

    :param str name: the name of the statistic.

    :returns str: the description of the statistic. If not found, returns an empty string
    """
    if not isinstance(name, str):
        raise TypeError("Statistic's name should be a string.")

    if name in profiles:
        return profiles[name]

    if name in "mean_trend10_zscore":
        return "Significance of (rolling) trend in means of features"

    head, *tails = name.split("_")
    tail = "_".join(tails)

    if tail in comparisons and head in references:
        return comparisons[tail]

    return ""


class SectionGenerator(Module):
    """This module takes the time-series data of already computed statistics, plots the data and
    combines all the plots into a list which is stored together with the section name in a dictionary
    which later will be used for the report generation.
    """

    _input_keys = ("read_key", "static_bounds", "dynamic_bounds", "store_key")
    _output_keys = ("store_key",)

    def __init__(
        self,
        read_key,
        store_key,
        section_name,
        settings: Report,
        features=None,
        ignore_features=None,
        static_bounds=None,
        dynamic_bounds=None,
        prefix: str = "traffic_light_",
        suffices=None,
        ignore_stat_endswith=None,
        description: str = "",
    ) -> None:
        """Initialize an instance of SectionGenerator.

        :param str read_key: key of input data to read from the datastore and use for plotting
        :param str store_key: key for output data to be stored in the datastore
        :param str section_name: key of output data to store in the datastore
        :param list features: list of features to pick up from input data (optional)
        :param list ignore_features: ignore list of features, if present (optional)
        :param str static_bounds: key to static traffic light bounds key in datastore (optional)
        :param str dynamic_bounds: key to dynamic traffic light bounds key in datastore (optional)
        :param str prefix: dynamic traffic light prefix. default is ``'traffic_light_'`` (optional)
        :param str suffices: dynamic traffic light suffices. (optional)
        :param list ignore_stat_endswith: ignore stats ending with any of list of suffices. (optional)
        :param str description: description of the section. default is empty (optional)
        """
        super().__init__()
        self.read_key = read_key
        self.store_key = store_key
        self.dynamic_bounds = dynamic_bounds
        self.static_bounds = static_bounds

        self.features = features or []
        self.ignore_features = ignore_features or []
        self.section_name = section_name
        self.last_n = settings.last_n
        self.skip_first_n = settings.skip_first_n
        self.skip_last_n = settings.skip_last_n
        self.prefix = prefix
        self.suffices = suffices or [
            "_red_high",
            "_yellow_high",
            "_yellow_low",
            "_red_low",
        ]
        self.ignore_stat_endswith = ignore_stat_endswith or []
        self.description = description
        self.show_stats = settings.show_stats if not settings.extended_report else None
        self.primary_color = settings.primary_color
        self.tl_colors = settings.tl_colors

    def get_description(self):
        return self.section_name

    def transform(
        self,
        data_obj: dict,
        static_bounds: dict | None = None,
        dynamic_bounds: dict | None = None,
        sections: list | None = None,
    ):
        if static_bounds is None:
            static_bounds = {}
        if dynamic_bounds is None:
            dynamic_bounds = {}
        if sections is None:
            sections = []

        features = self.get_features(list(data_obj.keys()))
        features_w_metrics = []

        self.logger.info(f'Generating section "{self.section_name}"')

        for feature in tqdm(features, ncols=100):
            df = data_obj.get(feature, pd.DataFrame())
            fdbounds = dynamic_bounds.get(feature, pd.DataFrame(index=df.index))

            assert all(df.index == fdbounds.index)

            # prepare date labels
            df = df.drop(
                columns=["histogram", "reference_histogram"],
                errors="ignore",
            )
            dates = np.array([short_date(date) for date in df.index.tolist()])

            metrics = filter_metrics(
                df.columns, self.ignore_stat_endswith, self.show_stats
            )

            args = [
                (
                    feature,
                    metric,
                    dates,
                    df[metric].values,
                    static_bounds,
                    fdbounds,
                    self.prefix,
                    self.suffices,
                    self.last_n,
                    self.skip_first_n,
                    self.skip_last_n,
                    self.primary_color,
                    self.tl_colors,
                )
                for metric in metrics
            ]
            plots = parallel(_plot_metric, args)

            # filter out potential empty plots (from skip empty plots)
            plots = [e for e in plots if len(e["plot"])]

            layouts = ""
            if len(plots) > 0:
                layouts = plots[0]["layout"]
                if "shapes" in layouts:
                    del layouts["shapes"]
                if "range" in layouts.get("yaxis", {}):
                    del layouts["yaxis"]["range"]

            # Group comparisons in Comparison section
            if self.read_key == "comparisons":
                grouped_metrics = defaultdict(list)
                for plot in plots:
                    prefix = plot["name"].split("_")[0]
                    if prefix not in references:
                        prefix = "Others"
                    grouped_metrics[prefix].append(plot)

                features_w_metrics.append(
                    {
                        "name": feature,
                        "plot_type_layouts": {"barplot": layouts},
                        "plots": grouped_metrics,
                        "titles": group_titles,
                        "descriptions": group_descriptions,
                    }
                )
            else:
                features_w_metrics.append(
                    {
                        "name": feature,
                        "plot_type_layouts": {"barplot": layouts},
                        "plots": plots,
                    }
                )

        sections.append(
            {
                "section_title": self.section_name,
                "section_description": self.description,
                "features": features_w_metrics,
            }
        )
        return sections


def _plot_metric(
    feature,
    metric,
    dates,
    values,
    static_bounds,
    fdbounds,
    prefix,
    suffices,
    last_n,
    skip_first_n,
    skip_last_n,
    primary_color,
    zline_color,
):
    """Split off plot histogram generation to allow for parallel processing"""
    # pick up static traffic light boundaries
    name = f"{feature}:{metric}"
    sbounds = static_bounds.get(name, ())
    # pick up dynamic traffic light boundaries
    names = [prefix + metric + suffix for suffix in suffices]
    dbounds = tuple(
        _prune(fdbounds[n].tolist(), last_n, skip_first_n, skip_last_n)
        for n in names
        if n in fdbounds.columns
    )
    # choose dynamic bounds if present
    bounds = dbounds if len(dbounds) > 0 else sbounds

    # prune dates and values
    dates = _prune(dates, last_n, skip_first_n, skip_last_n)
    values = _prune(values, last_n, skip_first_n, skip_last_n)

    # make plot. note: slow!
    plot = plot_bars(
        data=values,
        labels=dates,
        ylim=True,
        bounds=bounds,
        primary_color=primary_color,
        tl_colors=zline_color,
        metric=metric,
    )

    if not isinstance(plot, dict):
        return {
            "name": metric,
            "type": "barplot",
            "description": get_stat_description(metric),
            "plot": plot,
            "layout": plot,
        }

    return {
        "name": metric,
        "type": "barplot",
        "description": get_stat_description(metric),
        "plot": plot["data"],
        "shapes": plot["layout"]["shapes"] if "shapes" in plot["layout"] else "",
        "yaxis_range": (
            [
                "null" if r is None else r
                for r in plot["layout"].get("yaxis", {}).get("range")
            ]
            if "range" in plot["layout"].get("yaxis", {})
            else ""
        ),
        "layout": plot["layout"],
    }



# Copyright (c) 2023 ING Analytics Wholesale Banking
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import numpy as np
import pandas as pd

from popmon.analysis.apply_func import ApplyFunc
from popmon.analysis.functions import (
    expand_norm_hist_mean_cov,
    expanding_hist,
    hist_sum,
    normalized_hist_mean_cov,
    relative_chi_squared,
    roll_norm_hist_mean_cov,
    rolling_hist,
)
from popmon.analysis.hist_numpy import (
    check_similar_hists,
    get_consistent_numpy_entries,
    get_consistent_numpy_ndgrids,
)
from popmon.base import Pipeline
from popmon.hist.hist_utils import COMMON_HIST_TYPES, is_numeric


def hist_compare(row, hist_name1: str = "", hist_name2: str = ""):
    """Function to compare two histograms

    Apply statistical tests to compare two input histograms, such as:
    Chi2, KS, Pearson, max probability difference.
    For categorical histograms, also check for unknown labels.

    :param pd.Series row: row to apply compare function to
    :param str hist_name1: name of histogram one to compare
    :param str hist_name2: name of histogram two to compare
    :return: pandas Series with popular comparison metrics.
    """
    from popmon.analysis.comparison import Comparisons

    x = {key: np.nan for key in Comparisons.get_keys()}

    # basic name checks
    cols = row.index.to_list()
    if len(hist_name1) == 0 or len(hist_name2) == 0 and len(cols) == 2:
        hist_name1 = cols[0]
        hist_name2 = cols[1]
    if not all(name in cols for name in [hist_name1, hist_name2]):
        raise ValueError("Need to provide two histogram column names.")

    # basic histogram checks
    hist1 = row[hist_name1]
    hist2 = row[hist_name2]
    if not all(
        isinstance(hist, COMMON_HIST_TYPES) for hist in [hist1, hist2]
    ) or not check_similar_hists([hist1, hist2]):
        return pd.Series(x)

    # compare
    if hist1.n_dim == 1:
        entries_list = get_consistent_numpy_entries([hist1, hist2])
        if is_numeric(hist1):
            htype = "num"
            args = entries_list
        else:
            htype = "cat"
            args = [hist1, hist2]

        x.update(Comparisons.run(args, dim=1, htype=htype))
        x.update(Comparisons.run(entries_list, dim=1, htype="all"))
    else:
        numpy_ndgrids = get_consistent_numpy_ndgrids([hist1, hist2], dim=hist1.n_dim)
        entries_list = [entry.flatten() for entry in numpy_ndgrids]

        x.update(Comparisons.run(entries_list, dim=(2,), htype="all"))

    x.update(Comparisons.run(entries_list, dim=-1, htype="all"))

    if len(set(x.keys()) - set(Comparisons.get_keys())) > 0:
        raise ValueError("Could not compute full comparison")

    return pd.Series(x)


class HistComparer(Pipeline):
    """Base pipeline to compare histogram to previous rolling histograms"""

    def __init__(
        self,
        func_hist_collector,
        read_key,
        store_key,
        assign_to_key=None,
        hist_col: str = "histogram",
        suffix: str = "comp",
        *args,
        **kwargs,
    ) -> None:
        """Initialize an instance of RollingHistComparer.

        :param func_hist_collector: histogram collection function
        :param str read_key: key of input data to read from data store
        :param str store_key: key of output data to store in data store
        :param str assign_to_key: key of the input data to assign function applied-output to. (optional)
        :param str hist_col: column/key in input df/dict that contains the histogram. default is 'histogram'
        :param str suffix: column/key of rolling histogram. default is 'roll' -> column = 'histogram_roll'
        :param args: (tuple, optional): residual args passed on to func_mean and func_std
        :param kwargs: (dict, optional): residual kwargs passed on to func_mean and func_std
        """
        if assign_to_key is None:
            assign_to_key = read_key

        # make reference histogram(s)
        hist_collector = ApplyFunc(
            apply_to_key=read_key,
            assign_to_key=assign_to_key,
        )
        hist_collector.add_apply_func(
            *args, func=func_hist_collector, entire=True, suffix=suffix, **kwargs
        )
        # do histogram comparison
        hist_comparer = ApplyFunc(
            apply_to_key=assign_to_key,
            assign_to_key=store_key,
            apply_funcs=[
                {
                    "func": hist_compare,
                    "hist_name1": hist_col,
                    "hist_name2": hist_col + "_" + suffix,
                    "prefix": suffix,
                    "axis": 1,
                }
            ],
        )

        super().__init__(modules=[hist_collector, hist_comparer])


class RollingHistComparer(HistComparer):
    """Compare histogram to previous rolling histograms"""

    def __init__(
        self,
        read_key,
        store_key,
        window,
        shift: int = 1,
        hist_col: str = "histogram",
        suffix: str = "roll",
    ) -> None:
        """Initialize an instance of RollingHistComparer.

        :param str read_key: key of input data to read from data store
        :param str store_key: key of output data to store in data store
        :param int window: size of rolling window
        :param int shift: shift of rolling window. default is 1.
        :param str hist_col: column/key in input df/dict that contains the histogram. default is 'histogram'
        :param str suffix: column/key of rolling histogram. default is 'roll' -> column = 'histogram_roll'
        """
        super().__init__(
            rolling_hist,
            read_key,
            store_key,
            read_key,
            hist_col,
            suffix,
            window=window,
            shift=shift,
            hist_name=hist_col,
        )
        self.read_key = read_key
        self.window = window

    def transform(self, datastore):
        self.logger.info(
            f'Comparing "{self.read_key}" with rolling sum of {self.window} previous histogram(s).'
        )
        return super().transform(datastore)


class PreviousHistComparer(RollingHistComparer):
    """Compare histogram to previous histograms"""

    def __init__(
        self,
        read_key,
        store_key,
        hist_col: str = "histogram",
        suffix: str = "prev1",
    ) -> None:
        """Initialize an instance of PreviousHistComparer.

        :param str read_key: key of input data to read from data store
        :param str store_key: key of output data to store in data store
        :param str hist_col: column/key in input df/dict that contains the histogram. default is 'histogram'
        :param str suffix: column/key of rolling histogram. default is 'prev' -> column = 'histogram_prev'
        """
        super().__init__(
            read_key,
            store_key,
            window=1,
            shift=1,
            hist_col=hist_col,
            suffix=suffix,
        )


class ExpandingHistComparer(HistComparer):
    """Compare histogram to previous expanding histograms"""

    def __init__(
        self,
        read_key,
        store_key,
        shift: int = 1,
        hist_col: str = "histogram",
        suffix: str = "expanding",
    ) -> None:
        """Initialize an instance of ExpandingHistComparer.

        :param str read_key: key of input data to read from data store
        :param str store_key: key of output data to store in data store
        :param int shift: shift of rolling window. default is 1.
        :param str hist_col: column/key in input df/dict that contains the histogram. default is 'histogram'
        :param str suffix: column/key of rolling histogram. default is 'expanding' -> column = 'histogram_expanding'
        """
        super().__init__(
            expanding_hist,
            read_key,
            store_key,
            read_key,
            hist_col,
            suffix,
            shift=shift,
            hist_name=hist_col,
        )
        self.read_key = read_key

    def transform(self, datastore):
        self.logger.info(
            f'Comparing "{self.read_key}" with expanding sum of past histograms.'
        )
        return super().transform(datastore)


class ReferenceHistComparer(HistComparer):
    """Compare histogram to reference histograms"""

    def __init__(
        self,
        reference_key,
        assign_to_key,
        store_key,
        hist_col: str = "histogram",
        suffix: str = "ref",
    ) -> None:
        """Initialize an instance of ReferenceHistComparer.

        :param str reference_key: key of input data to read from data store
        :param str assign_to_key: key of input data to read from data store
        :param str store_key: key of output data to store in data store
        :param str hist_col: column/key in input df/dict that contains the histogram. default is 'histogram'
        :param str suffix: column/key of rolling histogram. default is 'ref' -> column = 'histogram_ref'
        """
        super().__init__(
            hist_sum,
            reference_key,
            store_key,
            assign_to_key,
            hist_col,
            suffix,
            metrics=[hist_col],
        )
        self.reference_key = reference_key
        self.assign_to_key = assign_to_key

    def transform(self, datastore):
        self.logger.info(
            f'Comparing "{self.assign_to_key}" with reference "{self.reference_key}"'
        )
        return super().transform(datastore)


class NormHistComparer(Pipeline):
    """Base pipeline to compare histogram to normalized histograms"""

    def __init__(
        self,
        func_hist_collector,
        read_key,
        store_key,
        assign_to_key=None,
        hist_col: str = "histogram",
        *args,
        **kwargs,
    ) -> None:
        """Initialize an instance of NormHistComparer.

        :param func_hist_collector: histogram collection function
        :param str read_key: key of input data to read from data store
        :param str store_key: key of output data to store in data store
        :param str assign_to_key: key of the input data to assign function applied-output to. (optional)
        :param str hist_col: column/key in input df/dict that contains the histogram. default is 'histogram'
        :param args: (tuple, optional): residual args passed on to func_hist_collector
        :param kwargs: (dict, optional): residual kwargs passed on to func_hist_collector
        """
        if assign_to_key is None:
            assign_to_key = read_key

        # make reference histogram(s)
        hist_collector = ApplyFunc(apply_to_key=read_key, assign_to_key=assign_to_key)
        hist_collector.add_apply_func(
            *args, func=func_hist_collector, hist_name=hist_col, suffix="", **kwargs
        )

        # do histogram comparison
        hist_comparer = ApplyFunc(
            apply_to_key=assign_to_key,
            assign_to_key=store_key,
            apply_funcs=[
                {
                    "func": relative_chi_squared,
                    "hist_name": hist_col,
                    "suffix": "",
                    "axis": 1,
                }
            ],
        )

        super().__init__(modules=[hist_collector, hist_comparer])


class RollingNormHistComparer(NormHistComparer):
    """Compare histogram to previous rolling normalized histograms"""

    def __init__(
        self, read_key, store_key, window, shift: int = 1, hist_col: str = "histogram"
    ) -> None:
        """Initialize an instance of RollingNormHistComparer.

        :param str read_key: key of input data to read from data store
        :param str store_key: key of output data to store in data store
        :param int window: size of rolling window
        :param int shift: shift of rolling window. default is 1.
        :param str hist_col: column/key in input df/dict that contains the histogram. default is 'histogram'
        """
        if window < 2:
            raise ValueError("Need window size of 2 or greater.")
        super().__init__(
            roll_norm_hist_mean_cov,
            read_key,
            store_key,
            read_key,
            hist_col,
            window=window,
            shift=shift,
            entire=True,
        )
        self.read_key = read_key
        self.window = window

    def transform(self, datastore):
        self.logger.info(
            f'Comparing "{self.read_key}" with relative mean of {self.window} previous histogram(s).'
        )
        return super().transform(datastore)


class ExpandingNormHistComparer(NormHistComparer):
    """Compare histogram to previous expanding normalized histograms"""

    def __init__(
        self, read_key, store_key, shift: int = 1, hist_col: str = "histogram"
    ) -> None:
        """Initialize an instance of ExpandingNormHistComparer.

        :param str read_key: key of input data to read from data store
        :param str store_key: key of output data to store in data store
        :param int shift: shift of rolling window. default is 1.
        :param str hist_col: column/key in input df/dict that contains the histogram. default is 'histogram'
        """
        super().__init__(
            expand_norm_hist_mean_cov,
            read_key,
            store_key,
            read_key,
            hist_col,
            shift=shift,
            entire=True,
        )
        self.read_key = read_key

    def transform(self, datastore):
        self.logger.info(
            f'Comparing "{self.read_key}" with normalized mean of expanding past histograms.'
        )
        return super().transform(datastore)


class ReferenceNormHistComparer(NormHistComparer):
    """Compare histogram to reference normalized histograms"""

    def __init__(
        self, reference_key, assign_to_key, store_key, hist_col: str = "histogram"
    ) -> None:
        """Initialize an instance of ReferenceNormHistComparer.

        :param str reference_key: key of input data to read from data store
        :param str assign_to_key: key of input data to read from data store
        :param str store_key: key of output data to store in data store
        :param str hist_col: column/key in input df/dict that contains the histogram. default is 'histogram'
        """
        super().__init__(
            normalized_hist_mean_cov, reference_key, store_key, assign_to_key, hist_col
        )
        self.reference_key = reference_key
        self.assign_to_key = assign_to_key

    def transform(self, datastore):
        self.logger.info(
            f'Comparing "{self.assign_to_key}" with normalized reference "{self.reference_key}"'
        )
        return super().transform(datastore)
